@@ -114,11 +114,10 @@ pager_class_init (YelpPagerClass *klass)
     g_object_class_install_property
 	(object_class,
 	 PROP_URI,
-	 g_param_spec_object ("uri",
-			      _("Document URI"),
-			      _("The URI of the document to be processed"),
-			      YELP_TYPE_URI,
-			      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE));
+	 g_param_spec_pointer ("uri",
+			       _("Document URI"),
+			       _("The URI of the document to be processed"),
+			       G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE));
 
     signals[START] = g_signal_new
 	("start",
@@ -203,12 +202,12 @@ pager_set_property (GObject      *object,
     switch (prop_id) {
     case PROP_URI:
 	if (pager->priv->uri)
-	    g_object_unref (pager->priv->uri);
+	    yelp_uri_unref (pager->priv->uri);
 
-	pager->priv->uri = (YelpURI *) g_value_get_object (value);
+	pager->priv->uri = (YelpURI *) g_value_get_pointer (value);
 
 	if (pager->priv->uri)
-	    g_object_ref (pager->priv->uri);
+	    yelp_uri_ref (pager->priv->uri);
 
 	break;
     default:
@@ -226,7 +225,7 @@ pager_get_property (GObject      *object,
 
     switch (prop_id) {
     case PROP_URI:
-	g_value_set_object (value, pager->priv->uri);
+	g_value_set_pointer (value, pager->priv->uri);
 	break;
     default:
 	break;
@@ -238,7 +237,7 @@ pager_dispose (GObject *object)
 {
     YelpPager *pager = YELP_PAGER (object);
 
-    g_object_unref (pager->priv->uri);
+    yelp_uri_unref (pager->priv->uri);
 
     if (pager->priv->error)
 	g_error_free (pager->priv->error);
@@ -368,7 +367,7 @@ yelp_pager_get_sections (YelpPager *pager)
 gchar *
 yelp_pager_resolve_uri (YelpPager *pager, YelpURI *uri)
 {
-    gchar    *frag_id = NULL;
+    gchar *frag_id = NULL;
 
     frag_id = (gchar *) (YELP_PAGER_GET_CLASS (pager)->resolve_uri (pager, uri));
 
@@ -412,9 +411,7 @@ yelp_pager_lookup_page (YelpPager *pager, YelpURI *uri)
     page_id = (gchar *) (YELP_PAGER_GET_CLASS (pager)->resolve_uri (pager, uri));
 
     if (!page_id)
-	page_id = yelp_uri_get_fragment (uri);
-    else
-	page_id = g_strdup (page_id);
+	page_id = g_strdup (gnome_vfs_uri_get_fragment_identifier (uri->uri));
 
     page = (YelpPage *) yelp_pager_get_page (pager, page_id);
 
