@@ -33,6 +33,7 @@
 
 #include "yelp-uri.h"
 #include "yelp-error.h"
+#include "yelp-theme.h"
 #include "yelp-toc-pager.h"
 
 #define d(x)
@@ -145,6 +146,8 @@ static gboolean      toc_process_idx_pending   (YelpTocPager      *pager);
 #endif
 
 static gchar *       menu_write_page           (YelpMenu          *menu);
+static void          menu_write_menu           (YelpMenu          *menu,
+						GString           *gstr);
 
 static xmlChar *     node_get_title            (xmlNodePtr         node);
 static void          toc_hash_metafile         (YelpTocPager      *pager,
@@ -231,8 +234,9 @@ void
 yelp_toc_pager_init (void)
 {
     YelpURI *uri;
-    
+
     uri = yelp_uri_new ("toc:");
+    
     toc_pager = (YelpTocPager *) g_object_new (YELP_TYPE_TOC_PAGER, 
     					       "uri", uri, NULL);
 
@@ -248,12 +252,14 @@ yelp_toc_pager_get (void)
 void
 yelp_toc_pager_pause (YelpTocPager *pager)
 {
+    g_return_if_fail (pager != NULL);
     pager->priv->pause_count = pager->priv->pause_count + 1;
 }
 
 void
 yelp_toc_pager_unpause (YelpTocPager *pager)
 {
+    g_return_if_fail (pager != NULL);
     pager->priv->pause_count = pager->priv->pause_count - 1;
 
     if (pager->priv->pause_count < 0) {
@@ -273,11 +279,35 @@ yelp_toc_pager_unpause (YelpTocPager *pager)
 gboolean
 toc_pager_process (YelpPager *pager)
 {
+    gchar  *manpath;
     YelpTocPagerPriv *priv = YELP_TOC_PAGER (pager)->priv;
 
+    /* Set the OMF directories to be read */
     priv->omf_dir_pending = g_slist_prepend (priv->omf_dir_pending,
 					     g_strdup (DATADIR "/omf"));
 
+    /* Set the man directories to be read */
+    if (!g_spawn_command_line_sync ("manpath", &manpath, NULL, NULL, NULL))
+	manpath = g_strdup (g_getenv ("MANPATH"));
+
+    if (manpath) {
+	gint    i;
+	gchar **paths;
+
+	g_strstrip (manpath);
+	paths = g_strsplit (manpath, G_SEARCHPATH_SEPARATOR_S, -1);
+                
+	for (i = 0; paths[i]; ++i)
+	    priv->man_dir_pending = g_slist_prepend (priv->man_dir_pending,
+						     g_strdup (paths[i]));
+
+	g_free (paths);
+	g_free (manpath);
+    }
+
+    /* Set the info directories to be read */
+
+    /* Set it running */
     priv->pending_func = (ProcessFunction) process_omf_dir_pending;
 
     gtk_idle_add_priority (G_PRIORITY_LOW,
@@ -297,16 +327,12 @@ toc_pager_cancel (YelpPager *pager)
 gchar *
 toc_pager_resolve_uri (YelpPager *pager, YelpURI *uri)
 {
-    gchar *path = yelp_uri_get_path (uri);
+    const gchar *frag = gnome_vfs_uri_get_fragment_identifier (uri->uri);
 
-    if (!strcmp (path, "")) {
-	g_free (path);
-	return g_strdup ("index");
-    }
-    else if (!path)
+    if (!frag)
 	return g_strdup ("index");
     else
-	return path;
+	return g_strdup (frag);
 }
 
 const GtkTreeModel *
@@ -622,10 +648,7 @@ process_read_menu (YelpTocPager *pager)
 				     XML_PARSE_NONET    );
 
     if (!priv->toc_doc) {
-	g_set_error (&error,
-		     YELP_ERROR,
-		     YELP_ERROR_FAILED_TOC,
-		     _("The table of contents could not be read."));
+	yelp_set_error (&error, YELP_ERROR_NO_TOC);
 	yelp_pager_error (YELP_PAGER (pager), error);
 	priv->cancel = TRUE;
 	return FALSE;
@@ -634,10 +657,7 @@ process_read_menu (YelpTocPager *pager)
     toc_node = xmlDocGetRootElement (priv->toc_doc);
 
     if (xmlStrcmp (toc_node->name, (const xmlChar *) "toc")) {
-	g_set_error (&error,
-		     YELP_ERROR,
-		     YELP_ERROR_FAILED_TOC,
-		     _("The table of contents could not be read."));
+	yelp_set_error (&error, YELP_ERROR_NO_TOC);
 	yelp_pager_error (YELP_PAGER (pager), error);
 	priv->cancel = TRUE;
 	return FALSE;
@@ -791,7 +811,13 @@ toc_process_idx_pending (YelpTocPager *pager)
     meta = first->data;
 
     uri  = yelp_uri_new (omf->file);
-    path = yelp_uri_get_path (uri);
+    path = gnome_vfs_uri_to_string (uri->uri,
+				    GNOME_VFS_URI_HIDE_USER_NAME           |
+				    GNOME_VFS_URI_HIDE_PASSWORD            |
+				    GNOME_VFS_URI_HIDE_HOST_NAME           |
+				    GNOME_VFS_URI_HIDE_HOST_PORT           |
+				    GNOME_VFS_URI_HIDE_TOPLEVEL_METHOD     |
+				    GNOME_VFS_URI_HIDE_FRAGMENT_IDENTIFIER );
 
     /*
     doc = xmlCtxtReadFile (priv->parser,
@@ -802,7 +828,7 @@ toc_process_idx_pending (YelpTocPager *pager)
     */
 
     g_free (path);
-    g_object_unref (uri);
+    yelp_uri_unref (uri);
     g_slist_free_1 (first);
 
     if (!priv->idx_pending) {
@@ -821,77 +847,60 @@ toc_process_idx_pending (YelpTocPager *pager)
 static gchar *
 menu_write_page (YelpMenu  *menu)
 {
-    gint      i = 0;
-    GSList   *c;
-    gchar   **strs;
-    gint      strs_len;
+    GString  *gstr;
     gchar    *page;
-    gint      page_len = 0;
-    gchar    *page_end;
 
-    strs_len =
-	g_slist_length (menu->submenus) + g_slist_length (menu->metafiles) + 20;
+    gstr = g_string_new ("<html>");
 
-    strs = g_new0 (gchar *, strs_len);
+    g_string_append_printf
+	(gstr,
+	 "<head><meta http-equiv='Content-Type'"
+	 " content='text/html=; charset=utf-8'>"
+	 "<link rel='stylesheet' type='text/css' href='%s'></head>\n",
+	 yelp_theme_get_css_file ());
 
-    strs[i] = g_strconcat ("<html><body><h1>", menu->title, "</h1>\n", NULL);
-    page_len += strlen (strs[i]);
+    g_string_append_printf (gstr, "<body><h1>%s</h1>\n", menu->title);
 
+    menu_write_menu (menu, gstr);
+
+    g_string_append (gstr, "</body></html>\n");
+
+    page = gstr->str;
+    g_string_free (gstr, FALSE);
+
+    return page;
+}
+
+static void
+menu_write_menu (YelpMenu  *menu,
+		 GString   *gstr)
+{
+    GSList   *c;
     if (menu->has_submenus) {
-	strs[++i] = g_strconcat ("<h2>", _("Categories"), "</h2><ul>\n", NULL);
-	page_len += strlen (strs[i]);
-
+	g_string_append_printf (gstr, "<h2>%s</h2><ul>\n", _("Categories"));
 	for (c = menu->submenus; c; c = c->next) {
 	    xmlNodePtr  cur      = (xmlNodePtr) c->data;
 	    YelpMenu   *cur_menu = (YelpMenu *) cur->_private;
-
-	    if (cur_menu && (cur_menu->has_submenus || cur_menu->metafiles)) {
-		strs[++i] = g_strconcat
-		    ("<li><a href='toc:", cur_menu->id, "'>",
-		     cur_menu->title,
-		     "</a></li>\n",
-		     NULL);
-		page_len += strlen (strs[i]);
-	    }
+	    if (cur_menu && (cur_menu->has_submenus || cur_menu->metafiles))
+		g_string_append_printf (gstr,
+					"<li><a href='toc:%s'>%s</a></li>\n",
+					cur_menu->id,
+					cur_menu->title);
 	}
-	strs[++i] = g_strdup ("</ul>\n");
-	page_len += strlen (strs[i]);
+	g_string_append (gstr, "</ul>\n");
     }
 
     if (menu->metafiles) {
-	strs[++i] = g_strconcat ("<h2>", _("Documents"), "</h2><ul>\n", NULL);
-	page_len += strlen (strs[i]);
-
+	g_string_append_printf (gstr, "<h2>%s</h2><ul>\n", _("Documents"));
 	for (c = menu->metafiles; c; c = c->next) {
 	    YelpMetafile *meta = (YelpMetafile *) c->data;
-	    strs[++i] = g_strconcat ("<li><a href='", meta->file, "'>",
-				     meta->title,
-				     "</a></li>\n",
-				     NULL);
-	    page_len += strlen (strs[i]);
+	    g_string_append_printf (gstr,
+				    "<li><a href='%s'>%s</a></li>\n",
+				    meta->file,
+				    meta->title);
 	}
-	strs[++i] = g_strdup ("</ul>\n");
-	page_len += strlen (strs[i]);
+	g_string_append (gstr, "</ul>\n");
     }
-
-    strs[++i] = g_strdup ("</body></html>\n");
-    page_len += strlen (strs[i]);
-
-    // Not necessary, but I'm paranoid.
-    strs[++i] = NULL;
-
-    page = g_new0 (gchar, page_len + 1);
-    page_end = page;
-
-    for (i = 0; i < strs_len; i++)
-	if (strs[i]) {
-	    page_end = g_stpcpy (page_end, strs[i]);
-	    g_free (strs[i]);
-	}
-
-    g_free (strs);
-
-    return page;
 }
 
 static xmlChar *
