@@ -43,6 +43,7 @@
 #include "yelp-error.h"
 #include "yelp-history.h"
 #include "yelp-html.h"
+#include "yelp-man-pager.h"
 #include "yelp-pager.h"
 #include "yelp-toc-pager.h"
 #include "yelp-theme.h"
@@ -54,6 +55,8 @@
 #define YELP_CONFIG_WIDTH_DEFAULT  "600"
 #define YELP_CONFIG_HEIGHT_DEFAULT "420"
 
+#define BUFFER_SIZE 16384
+
 typedef enum {
     YELP_WINDOW_FIND_PREV = 1,
     YELP_WINDOW_FIND_NEXT
@@ -105,6 +108,9 @@ static void        html_uri_selected_cb           (YelpHtml          *html,
 						   YelpURI           *uri,
 						   gboolean           handled,
 						   gpointer           user_data);
+static void        html_title_changed_cb          (YelpHtml          *html,
+						   gchar             *title,
+						   gpointer           user_data);
 static void        tree_selection_changed_cb      (GtkTreeSelection  *selection,
 						   YelpWindow        *window);
 
@@ -250,10 +256,10 @@ static GtkItemFactoryEntry menu_items[] = {
      window_go_cb,                   YELP_WINDOW_GO_HOME,
      "<StockItem>",                  GTK_STOCK_HOME             },
     {N_("/Go/"),                     NULL, 0, 0, "<Separator>"  },
-    {N_("/Go/_Previous"),            "<Alt>Up",
+    {N_("/Go/_Previous Page"),       "<Alt>Up",
      window_go_cb,                   YELP_WINDOW_GO_PREVIOUS,
      0,                              NULL                       },
-    {N_("/Go/_Next"),                "<Alt>Down",
+    {N_("/Go/_Next Page"),           "<Alt>Down",
      window_go_cb,                   YELP_WINDOW_GO_NEXT,
      0,                              NULL                       },
     {N_("/Go/_Contents"),            NULL,
@@ -347,7 +353,7 @@ yelp_window_new (GNode *doc_tree, GList *index)
 {
     YelpWindow     *window;
     YelpWindowPriv *priv;
-	
+
     window = g_object_new (YELP_TYPE_WINDOW, NULL);
 
     priv   = window->priv;
@@ -367,35 +373,33 @@ yelp_window_open_uri (YelpWindow  *window,
 		      YelpURI     *uri)
 {
     YelpWindowPriv *priv;
-    GError         *error = NULL;
     YelpURI        *cur_uri;
 
     g_return_if_fail (YELP_IS_WINDOW (window));
-    g_return_if_fail (YELP_IS_URI (uri));
+
+    if (uri == NULL) {
+	GError *error = NULL;
+	yelp_set_error (&error, YELP_ERROR_NO_DOC);
+	window_error (window, error);
+	g_error_free (error);
+	return;
+    }
 
     priv = window->priv;
 
     cur_uri = yelp_history_get_current (window->priv->history);
-    if (cur_uri && yelp_uri_equal (cur_uri, uri))
- 	return;
+    if (cur_uri && gnome_vfs_uri_equal (cur_uri->uri, uri->uri)) {
+	const gchar *cur_frag, *frag;
+	cur_frag = gnome_vfs_uri_get_fragment_identifier (cur_uri->uri);
+	frag     = gnome_vfs_uri_get_fragment_identifier (uri->uri);
+
+	if ((!cur_frag && !frag)
+	    || (cur_frag && frag && !strcmp (cur_frag, frag)))
+	    return;
+    }
 
     yelp_history_goto (window->priv->history, uri);
 
-    if (!yelp_uri_exists (uri)) {
-	gchar *str_uri = yelp_uri_to_string (uri);
-		
-	g_set_error (&error,
-		     YELP_ERROR,
-		     YELP_ERROR_URI_NOT_EXIST,
-		     _("The document '%s' does not exist"), str_uri);
-	window_error (window, error);
-
-	g_free (str_uri);
-	g_error_free (error);
-
-	return;
-    }
-
     window_handle_uri (window, uri);
 }
 
@@ -406,9 +410,10 @@ window_handle_uri (YelpWindow  *window,
     YelpWindowPriv *priv;
     GError         *error = NULL;
     gboolean        handled = FALSE;
+    gchar          *str_uri;
 
     g_return_if_fail (YELP_IS_WINDOW (window));
-    g_return_if_fail (YELP_IS_URI (uri));
+    g_return_if_fail (uri != NULL);
 
     priv = window->priv;
 
@@ -420,30 +425,26 @@ window_handle_uri (YelpWindow  *window,
 	handled = window_handle_pager_uri (window, uri);
 	break;
     case YELP_URI_TYPE_DOCBOOK_SGML:
-	g_set_error (&error,
-		     YELP_ERROR,
-		     YELP_ERROR_NO_SGML,
-		     _("DocBook SGML documents are no longer supported."));
-
-	window_error (window, error);
-	g_error_free (error);
-	return;
+	yelp_set_error (&error, YELP_ERROR_NO_SGML);
+	break;
     case YELP_URI_TYPE_HTML:
 	handled = window_handle_html_uri (window, uri);
 	break;
-    case YELP_URI_TYPE_GHELP:
-    case YELP_URI_TYPE_GHELP_OTHER:
-    case YELP_URI_TYPE_INDEX:
-    case YELP_URI_TYPE_PATH:
-    case YELP_URI_TYPE_FILE:
-    case YELP_URI_TYPE_UNKNOWN:
-    case YELP_URI_TYPE_RELATIVE:
+    case YELP_URI_TYPE_EXTERNAL:
+	str_uri = gnome_vfs_uri_to_string (uri->uri, GNOME_VFS_URI_HIDE_NONE);
+	gnome_url_show (str_uri, &error);
+	g_free (str_uri);
+	break;
+    case YELP_URI_TYPE_ERROR:
     default:
+	yelp_set_error (&error, YELP_ERROR_NO_DOC);
 	break;
     }
  
-    if (error)
+    if (error) {
 	window_error (window, error);
+	g_error_free (error);
+    }
 
     window_find_buttons_set_sensitive (window, TRUE);    
 }
@@ -452,7 +453,7 @@ YelpURI *
 yelp_window_get_current_uri (YelpWindow *window)
 {
     g_return_val_if_fail (YELP_IS_WINDOW (window), NULL);
-	
+
     return yelp_history_get_current (window->priv->history);
 }
 
@@ -566,6 +567,10 @@ window_populate (YelpWindow *window)
 		      "uri_selected",
 		      G_CALLBACK (html_uri_selected_cb),
 		      window);
+    g_signal_connect (priv->html_view,
+		      "title_changed",
+		      G_CALLBACK (html_title_changed_cb),
+		      window);
 
     gtk_container_add (GTK_CONTAINER (priv->html_sw),
 		       yelp_html_get_widget (priv->html_view));
@@ -705,17 +710,16 @@ window_set_sections (YelpWindow   *window,
 }
 
 static gboolean
-window_handle_pager_uri (YelpWindow *window,
-			 YelpURI    *uri)
+window_handle_pager_uri (YelpWindow  *window,
+			 YelpURI     *uri)
 {
     YelpWindowPriv *priv;
     GError         *error = NULL;
-    gboolean    loadnow  = FALSE;
-    gboolean    startnow = TRUE;
-    gchar      *str_uri;
-    gchar      *path;
-    YelpPage   *page = NULL;
-    YelpPager  *pager;
+    gboolean     loadnow  = FALSE;
+    gboolean     startnow = TRUE;
+    gchar       *path;
+    YelpPage    *page = NULL;
+    YelpPager   *pager;
     YelpPagerState state;
 
     priv = window->priv;
@@ -726,7 +730,7 @@ window_handle_pager_uri (YelpWindow *window,
     if (yelp_uri_get_resource_type (uri) == YELP_URI_TYPE_TOC) {
 	pager = YELP_PAGER (yelp_toc_pager_get ());
     } else {
-	path  = yelp_uri_get_path (uri);
+	path  = g_strdup (gnome_vfs_uri_get_path (uri->uri));
 	pager = (YelpPager *) yelp_cache_lookup (path);
 
 	// Create a new pager if one doesn't exist in the cache
@@ -739,7 +743,7 @@ window_handle_pager_uri (YelpWindow *window,
 		// FIXME: yelp_info_pager_new (uri);
 		break;
 	    case YELP_URI_TYPE_MAN:
-		// FIXME: yelp_man_pager_new (uri);
+		pager = yelp_man_pager_new (uri);
 		break;
 	    default:
 		break;
@@ -751,13 +755,8 @@ window_handle_pager_uri (YelpWindow *window,
     }
 
     if (!pager) {
-	str_uri = yelp_uri_to_string (uri);
-	g_set_error (&error,
-		     YELP_ERROR,
-		     YELP_ERROR_FAILED_OPEN,
-		     _("The document '%s' could not be opened"), str_uri);
+	yelp_set_error (&error, YELP_ERROR_NO_DOC);
 	window_error (window, error);
-	g_free (str_uri);
 	g_error_free (error);
     }
 
@@ -781,25 +780,20 @@ window_handle_pager_uri (YelpWindow *window,
 
     if (state & YELP_PAGER_STATE_STARTED) {
 	if (state & YELP_PAGER_STATE_CONTENTS) {
-	    gchar *frag_id = yelp_uri_get_fragment (uri);
+	    const gchar *frag_id =
+		gnome_vfs_uri_get_fragment_identifier (uri->uri);
 	    gchar *page_id = yelp_pager_resolve_uri (pager, uri);
 
 	    if (!page_id && (frag_id && strcmp (frag_id, ""))) {
-		g_set_error (&error,
-			     YELP_ERROR,
-			     YELP_ERROR_FAILED_OPEN,
-			     _("The page '%s' could not be found in this document."),
-			     frag_id);
+		yelp_set_error (&error, YELP_ERROR_NO_PAGE);
 		window_error (window, error);
 
-		g_free (frag_id);
 		g_free (page_id);
 		g_error_free (error);
 
 		return FALSE;
 	    }
 
-	    g_free (frag_id);
 	    g_free (page_id);
 	} else {
 	    priv->contents_handler =
@@ -815,15 +809,8 @@ window_handle_pager_uri (YelpWindow *window,
 
 	if (state & YELP_PAGER_STATE_FINISHED) {
 	    if (!page) {
-		gchar *frag_id = yelp_uri_get_fragment (uri);
-		g_set_error (&error,
-			     YELP_ERROR,
-			     YELP_ERROR_FAILED_OPEN,
-			     _("The page '%s' could not be found in this document."),
-			     frag_id);
+		yelp_set_error (&error, YELP_ERROR_NO_PAGE);
 		window_error (window, error);
-
-		g_free (frag_id);
 		g_error_free (error);
 		return FALSE;
 	    }
@@ -913,13 +900,62 @@ static gboolean
 window_handle_html_uri (YelpWindow    *window,
 			YelpURI       *uri)
 {
-    return FALSE;
+    GnomeVFSHandle  *handle;
+    GnomeVFSResult   result;
+    GnomeVFSFileSize n;
+    gchar            buffer[BUFFER_SIZE];
+    GtkWidget       *menu_item;
+
+    YelpWindowPriv *priv = window->priv;
+
+    g_return_val_if_fail (YELP_IS_WINDOW (window), FALSE);
+    g_return_val_if_fail (uri != NULL, FALSE);
+
+    window_set_sections (window, NULL);
+
+    menu_item =
+	gtk_item_factory_get_item_by_action (priv->item_factory,
+					     YELP_WINDOW_GO_PREVIOUS);
+    if (menu_item)
+	gtk_widget_set_sensitive (menu_item, FALSE);
+    menu_item =
+	gtk_item_factory_get_item_by_action (priv->item_factory,
+					     YELP_WINDOW_GO_NEXT);
+    if (menu_item)
+	gtk_widget_set_sensitive (menu_item, FALSE);
+    menu_item =
+	gtk_item_factory_get_item_by_action (priv->item_factory,
+					     YELP_WINDOW_GO_TOC);
+    if (menu_item)
+	gtk_widget_set_sensitive (menu_item, FALSE);
+
+
+    yelp_html_clear (priv->html_view);
+    yelp_html_set_base_uri (priv->html_view, uri);
+
+    result = gnome_vfs_open (&handle,
+			     gnome_vfs_uri_get_path (uri->uri),
+			     GNOME_VFS_OPEN_READ);
+
+    if (result != GNOME_VFS_OK) {
+	// FIXME: Give an error
+	return FALSE;
+    }
+
+    while ((result = gnome_vfs_read
+	    (handle, buffer, BUFFER_SIZE, &n)) == GNOME_VFS_OK) {
+	yelp_html_write (priv->html_view, buffer, n);
+    }
+
+    gnome_vfs_close (handle);
+
+    return TRUE;
 }
 
 static void
 window_handle_page (YelpWindow   *window,
 		    YelpPage     *page,
-		    YelpURI      *uri)
+		    YelpURI  *uri)
 {
     GtkWidget      *menu_item;
     GtkTreeModel   *model;
@@ -929,7 +965,6 @@ window_handle_page (YelpWindow   *window,
     gboolean        valid;
 
     g_return_if_fail (YELP_IS_WINDOW (window));
-    g_return_if_fail (YELP_IS_URI (uri));
 
     priv = window->priv;
 
@@ -990,6 +1025,12 @@ window_handle_page (YelpWindow   *window,
     yelp_html_write (priv->html_view,
 		     page->chunk,
 		     strlen (page->chunk));
+
+    if (gnome_vfs_uri_get_fragment_identifier (uri->uri)) {
+	yelp_html_jump_to_anchor
+	    (priv->html_view,
+	     (gchar *) gnome_vfs_uri_get_fragment_identifier (uri->uri));
+    }
 }
 
 static void
@@ -1026,29 +1067,24 @@ static void
 pager_contents_cb (YelpPager   *pager,
 		   gpointer     user_data)
 {
-    YelpWindow *window = YELP_WINDOW (user_data);
-    YelpURI    *uri;
-    GError     *error = NULL;
+    YelpWindow  *window = YELP_WINDOW (user_data);
+    YelpURI     *uri    = NULL;
+    GError      *error  = NULL;
+    const gchar *frag;
+    gchar       *page;
 
     uri  = yelp_window_get_current_uri (window);
-
-    gchar *frag = yelp_uri_get_fragment (uri);
-    gchar *page = yelp_pager_resolve_uri (pager, uri);
+    frag = gnome_vfs_uri_get_fragment_identifier (uri->uri);
+    page = yelp_pager_resolve_uri (pager, uri);
 
     if (!page && (frag && strcmp (frag, ""))) {
-	gchar *frag_id = yelp_uri_get_fragment (uri);
-	g_set_error (&error,
-		     YELP_ERROR,
-		     YELP_ERROR_FAILED_OPEN,
-		     _("The page '%s' could not be found in this document."),
-		     frag_id);
-	window_error (window, error);
-
-	g_free (frag_id);
-	g_error_free (error);
-
 	window_disconnect (window);
+
+	yelp_set_error (&error, YELP_ERROR_NO_PAGE);
+	window_error (window, error);
+	g_error_free (error);
     }
+    g_free (page);
 }
 
 static void
@@ -1056,9 +1092,9 @@ pager_page_cb (YelpPager *pager,
 	       gchar     *page_id,
 	       gpointer   user_data)
 {
-    YelpWindow *window = YELP_WINDOW (user_data);
-    YelpURI    *uri;
-    YelpPage   *page;
+    YelpWindow  *window = YELP_WINDOW (user_data);
+    YelpURI     *uri;
+    YelpPage    *page;
 
     uri  = yelp_window_get_current_uri (window);
 
@@ -1090,23 +1126,18 @@ pager_finish_cb (YelpPager   *pager,
 		 gpointer     user_data)
 {
     GError *error = NULL;
-    YelpWindow *window = YELP_WINDOW (user_data);
-    YelpURI *uri;
-    gchar   *frag_id;
+    YelpWindow  *window = YELP_WINDOW (user_data);
+    YelpURI     *uri;
+    const gchar *frag;
 
-    uri     = yelp_window_get_current_uri (window);
-    frag_id = yelp_uri_get_fragment (uri);
+    uri  = yelp_window_get_current_uri (window);
+    frag = gnome_vfs_uri_get_fragment_identifier (uri->uri);
 
     window_disconnect (window);
 
-    g_set_error (&error,
-		 YELP_ERROR,
-		 YELP_ERROR_FAILED_OPEN,
-		 _("The page '%s' could not be found in this document."),
-		 frag_id);
+    yelp_set_error (&error, YELP_ERROR_NO_PAGE);
     window_error (window, error);
 
-    g_free (frag_id);
     g_error_free (error);
 
     // FIXME: Remove the URI from the history and go back
@@ -1119,17 +1150,29 @@ html_uri_selected_cb (YelpHtml  *html,
 		      gpointer   user_data)
 {
     YelpWindow *window = YELP_WINDOW (user_data);
-    yelp_window_open_uri (window, uri);
+
+    if (!handled) {
+	yelp_window_open_uri (window, uri);
+	yelp_uri_unref (uri);
+    }
+}
+
+static void
+html_title_changed_cb (YelpHtml  *html,
+		       gchar     *title,
+		       gpointer   user_data)
+{
+    gtk_window_set_title (GTK_WINDOW (user_data), title);
 }
 
 static void
 tree_selection_changed_cb (GtkTreeSelection *selection,
 			   YelpWindow       *window)
 {
-    GtkTreeModel    *model;
-    GtkTreeIter      iter;
-    gchar           *id, *frag;
-    YelpURI         *uri;
+    GtkTreeModel  *model;
+    GtkTreeIter    iter;
+    gchar         *id, *frag;
+    YelpURI       *uri;
 
     YelpWindowPriv *priv = window->priv;
 
@@ -1141,7 +1184,7 @@ tree_selection_changed_cb (GtkTreeSelection *selection,
 			    -1);
 
 	frag = g_strconcat ("#", id, NULL);
-	uri = yelp_uri_new_relative
+	uri = yelp_uri_resolve_relative
 	    (yelp_window_get_current_uri (window), frag);
 
 	yelp_window_open_uri (window, uri);
@@ -1150,7 +1193,6 @@ tree_selection_changed_cb (GtkTreeSelection *selection,
     }
 }
 
-
 /******************************************************************************/
 
 static void
@@ -1242,7 +1284,7 @@ window_go_cb (gpointer   data,
 	      GtkWidget *widget)
 {
     YelpWindow *window;
-    YelpURI *uri;
+    YelpURI    *uri;
 
     g_return_if_fail (YELP_IS_WINDOW (YELP_WINDOW (data)));
 
@@ -1258,22 +1300,22 @@ window_go_cb (gpointer   data,
 	break;
 
     case YELP_WINDOW_GO_PREVIOUS:
-	uri = yelp_uri_new_relative (yelp_window_get_current_uri (window),
-				     window->priv->prev);
+	uri = yelp_uri_resolve_relative (yelp_window_get_current_uri (window),
+					 window->priv->prev);
 	yelp_window_open_uri (window, uri);
-	g_object_unref (uri);
+	yelp_uri_unref (uri);
 	break;
     case YELP_WINDOW_GO_NEXT:
-	uri = yelp_uri_new_relative (yelp_window_get_current_uri (window),
-				     window->priv->next);
+	uri = yelp_uri_resolve_relative (yelp_window_get_current_uri (window),
+					 window->priv->next);
 	yelp_window_open_uri (window, uri);
-	g_object_unref (uri);
+	yelp_uri_unref (uri);
 	break;
     case YELP_WINDOW_GO_TOC:
-	uri = yelp_uri_new_relative (yelp_window_get_current_uri (window),
-				     window->priv->toc);
+	uri = yelp_uri_resolve_relative (yelp_window_get_current_uri (window),
+					 window->priv->toc);
 	yelp_window_open_uri (window, uri);
-	g_object_unref (uri);
+	yelp_uri_unref (uri);
 	break;
     default:
 	break;
@@ -1288,11 +1330,9 @@ window_home_button_clicked (GtkWidget  *button,
 	
     g_return_if_fail (YELP_IS_WINDOW (window));
 	
-    uri = yelp_uri_new ("toc:");
+    uri = yelp_pager_get_uri (YELP_PAGER (yelp_toc_pager_get ()));
 
     yelp_window_open_uri (window, uri);
-
-    g_object_unref (uri);
 }
 
 static void
